from typing import Any, Callable, NoReturn, TypeVar

from typing_extensions import Protocol

# Single-sourcing the version number with poetry:
# https://github.com/python-poetry/poetry/pull/2366#issuecomment-652418094
try:
    __version__ = __import__("importlib.metadata").metadata.version(__name__)
except ModuleNotFoundError:  # pragma: no cover
    __version__ = __import__("importlib_metadata").version(__name__)


Q = TypeVar("Q")
R = TypeVar("R")
S = TypeVar("S")
T = TypeVar("T", contravariant=True)
U = TypeVar("U", contravariant=True)
V = TypeVar("V", covariant=True)


def raises(__e: BaseException) -> Callable[..., NoReturn]:
    """Create a callable which raises the given exception

    Example
    -------

    >>> f = raises(ValueError("foo"))
    ...
    >>> f()  # any arguments are accepted
    Traceback (most recent call last):
      File "<stdin>", line 1, in <module>
    ValueError: foo
    """

    def __raises(*_: Any, **__: Any) -> NoReturn:
        raise __e

    return __raises


def const(__v: S) -> Callable[..., S]:
    """Create a callable which always returns the same value

    Example
    -------

    >>> f = const(4)
    ...
    >>> f()  # any arguments are accepted
    4
    """
    return lambda *_, **__: __v


def identity(__v: S) -> S:
    """Pass the given value unmodified

    Example
    -------

    >>> identity(4)
    4
    >>> identity("foo")
    "foo"
    """
    return __v


def flip(__f: "BinaryFunction[Q, R, S]") -> "BinaryFunction[R, Q, S]":
    """Flip the two arguments of a function.
    Often useful in combination with :func:`~functools.partial`.

    Example
    -------

    >>> f = flip(round)
    ...
    >>> f(2, 5.125)
    5.13
    >>> f(0, 3.4)
    3
    """

    def __flipped(__a: R, __b: Q) -> S:
        return __f(__b, __a)

    return __flipped


class UnaryFunction(Protocol[T, V]):
    """A function/callable taking exactly one argument.

    Examples of objects satisfying this protocol:

    >>> # Below are all UnaryFunction[int, float]
    ...
    >>> def myfunc(a: int) -> float:
    ...     return a / 7
    ...
    >>> lambda a: a / 7
    ...
    >>> (7).__rtruediv__
    """

    def __call__(self, __value: T) -> V:
        """ """


class BinaryFunction(Protocol[T, U, V]):
    """A function/callable which takes exactly two arguments."""

    def __call__(self, __a: T, __b: U) -> V:
        """ """



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['calls']

package_data = \
{'': ['*']}

install_requires = \
['typing-extensions>=3.7.4,<4.0.0']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=1,<5']}

setup_kwargs = {
    'name': 'calls',
    'version': '0.1.1',
    'description': 'Utilities for callables',
    'long_description': "🤙 Calls\n========\n\n.. image:: https://img.shields.io/pypi/v/calls.svg?style=flat-square\n   :target: https://pypi.python.org/pypi/calls\n\n.. image:: https://img.shields.io/pypi/l/calls.svg?style=flat-square\n   :target: https://pypi.python.org/pypi/calls\n\n.. image:: https://img.shields.io/pypi/pyversions/calls.svg?style=flat-square\n   :target: https://pypi.python.org/pypi/calls\n\n.. image:: https://img.shields.io/readthedocs/calls.svg?style=flat-square\n   :target: http://calls.readthedocs.io/\n\n.. image:: https://img.shields.io/badge/code%20style-black-000000.svg?style=flat-square\n   :target: https://github.com/psf/black\n\nSimple, typed, composable tools for callables.\n\nQuickstart\n----------\n\nAll utilities are directly importable.\nSee `the docs <https://calls.rtfd.io>`_ for a complete overview.\n\n.. code-block:: python\n\n  >>> from calls import raises, const, identity, flip\n\nInstallation\n------------\n\nIt's available on PyPI.\n\n.. code-block:: bash\n\n  pip install calls\n",
    'author': 'Arie Bovenberg',
    'author_email': 'a.c.bovenberg@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ariebovenberg/calls',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
